#!/bin/bash
#
# CopyLeft moon.cat
# https://creativecommons.org/licenses/by/4.0/
#
# Author: Rei Izumi - moon.cat - 2021
#
# Simple utility to create a self-signed root CA and generate certificates.
#
# Exit codes:
# - 1: Argument does not contain value
# - 2: Missing argument (name of the certificate)
# - 3: Missing argument (root CA name)
# - 4: Missing argument (DNS)
#

####################
# Process          #
####################
readonly conf_file="req.conf"
readonly script_name="${0##*/}"

usage() {
    cat <<USAGE_TEXT
Usage: ${script_name} command [-h | --help] [<options>]
DESCRIPTION
	This script allows to generate the root CA certificate and any other signed by it.
	All certificate expire after 10 years.

COMMANDS:
create-ca
	Create the root CA certificate with KEY, CRT and PEM with encryption of 8192.
	This root CA does not have a password, so the KEY must be protected!
	Anyone with this KEY file can sign any certificate.
create-cert
	Create a new certificate signed by the root CA.
	This has the KEY and CRT and is encrypted with 4192.

OPTIONS:
-h, --help
        Print this help and exit.
-n, --name
        Name of the file. Same is used in all files.
	Required by create-ca and create-cert.
-r, --ca-name
        Name of the root CA file to sign the new certificate.
	Required by create-cert.
-s, --subj
        Values to indicate in the certificate.
	Example: '/C=ES/ST=Barcelona/L=Barcelona/O=Domain Cat/CN=*.domain.cat'
	Recommended by create-ca and required by create-cert.
	If not indicated, the original wizard will be displayed to fill it.
-d, --dns
	Additional DNS. Although the certificate only have one CN, these DNS are necessary to avoid incompatiblities.
	Example: DNS.1=domain.cat\\nDNS.2=*.domain.cat
	Required by create-cert.
USAGE_TEXT
exit 0
}

# Create config file if it does not exist
if [[ ! -f "${conf_file}" ]]; then
    cat > $conf_file << EOF
[req]
distinguished_name = req_distinguished_name
req_extensions = v3_req
prompt = no

[req_distinguished_name]

[CA_default]
copy_extensions = copy

[v3_req]
keyUsage = keyEncipherment, digitalSignature, nonRepudiation
extendedKeyUsage = serverAuth
subjectAltName = @alt_names
[alt_names]

EOF
fi

# Read commands and arguments
CREATE_CA=false;
CREATE_CERT=false;
while (( "$#" )); do
  case "$1" in
    -h|--help)
      usage
      shift
      ;;
    create-ca)
      CREATE_CA=true
      shift
      ;;
    create-cert)
      CREATE_CERT=true
      shift
      ;;

    -n|--name)
      if [ -n "$2" ] && [ ${2:0:1} != "-" ]; then
        NAME=$2
        shift 2
      else
        echo "Error: Argument for $1 is missing" >&2
        exit 1
      fi
      ;;
    -r|--ca-name)
      if [ -n "$2" ] && [ ${2:0:1} != "-" ]; then
        CANAME=$2
        shift 2
      else
        echo "Error: Argument for $1 is missing" >&2
        exit 1
      fi
      ;;
    -s|--subj)
      if [ -n "$2" ] && [ ${2:0:1} != "-" ]; then
        SUBJ=$2
        shift 2
      else
        echo "Error: Argument for $1 is missing" >&2
        exit 1
      fi
      ;;
    -d|--dns)
      if [ -n "$2" ] && [ ${2:0:1} != "-" ]; then
        DNS=$2
        shift 2
      else
        echo "Error: Argument for $1 is missing" >&2
        exit 1
      fi
      ;;
    *) # Discard others
      shift
      ;;
  esac
done

# If no command is selected, show the help
if [ "$CREATE_CA" = false -a "$CREATE_CERT" = false ]; then
  usage
fi

# Process to generate the root CA
if [ "$CREATE_CA" = true ]; then
  echo "Root CA will be generated";
  if [ -z "$NAME" ]; then
    echo "No name specified for the certificate (-n|--name)";
    exit 2
  fi
  
  echo "- Generating key"
  openssl genrsa -out $NAME.key 8196

  echo "- Generating crt"
  if [ -z "$SUBJ" ]; then
    openssl req -x509 -new -nodes -key $NAME.key -sha256 -days 3650 \
	-out $NAME.crt
  else
    openssl req -x509 -new -nodes -key $NAME.key -sha256 -days 3650 \
	-subj "$SUBJ" \
	-out $NAME.crt
  fi

  echo "- Generating pem"
  openssl x509 -in $NAME.crt -out $NAME.pem -outform PEM

# Process to generate and sign the new certificate
else
  echo "New certificate will be created"
  if [ -z "$NAME" ]; then
    echo "No name specified for the certificate (-n|--name)";
    exit 2
  fi
  if [ -z "$CANAME" ]; then
    echo "No root CA name specified to sign this new certificate (-r|--ca-name)";
    exit 3
  fi
  if [ -z "$DNS" ]; then
    echo "Alternative DNS must be specified (-d|--dns)";
    exit 4
  fi
  
  echo "- Generating key"
  openssl genrsa -out $NAME.key 4096

  echo "- Generating csr"
  if [ -z "$SUBJ" ]; then
    openssl req -new -sha256 -key $NAME.key \
	-out $NAME.csr \
	-config <(cat req.conf \
        <(printf "$DNS"))
  else
    openssl req -new -sha256 -key $NAME.key \
	-subj "$SUBJ" \
	-out $NAME.csr \
	-config <(cat req.conf \
        <(printf "$DNS"))
  fi

  echo "- Signing certificate"
  openssl x509 -req -in $NAME.csr \
	-CA $CANAME.crt -CAkey $CANAME.key -CAcreateserial \
	-out $NAME.crt -days 3650 -sha256 \
	-extensions v3_req \
	-extfile <(cat req.conf \
        <(printf "$DNS"))

fi

exit 0
